<?php
namespace Zeedhi\Framework\Service;

/**
 * Translator class.
 * @class
 *
 * A compile of all translation methods.
 */
class Translator {
    /**
     * Returns the translation file as array.
     *
     * @param string $dirPath   The directory path.
     * @param string $language  The target language.
     *
     * @return array    The translation file as array.
     */
    public static function getTranslations($dirPath, $language) : array {
        if(!$language){
            return [];
        }
        $filePath = realpath($dirPath.DIRECTORY_SEPARATOR."$language.json");
        $fileContent = self::getFileContent($filePath);
        $translation = self::parseToArray($fileContent);
        return $translation;
    }

    /**
     * Parses a stringified json to array.
     *
     * @param string $content The content as string.
     *
     * @return array $parsed The parsed content.
     *
     * @throws \Exception Can't read from file!
     */
    public static function parseToArray($content) : array {
        $parsed = json_decode($content, true);
        if(!$parsed){
            throw new \Exception(json_last_error_msg());
        }
        return $parsed;
    }

    /**
     * Returns a file content.
     *
     * @param string $filePath The file path.
     *
     * @return string $file The file content.
     *
     * @throws \Exception Can't read from file!
     */
    public static function getFileContent($filePath) : string {
        try{
            return file_get_contents($filePath);
        } catch(\Exception $e){
            throw new \Exception("Failed to fetch translation file $filePath. Error {$e->getCode()} with message: {$e->getMessage()}");
        }
    }

    /**
     * Replaces the variables from a content using a mapped variables array.
     *
     * @param array $content    The content as array.
     * @param array $content    The mapped variables array.
     *
     * @return array    The decoded content.
     */
    public static function replaceVariables(array $content, array $variables) : array {
        /** @var callable $walkCallback The content array walk callback. */
        $walkCallback = function(&$value) use ($variables) {
            /** @var callable $replaceCallback The preg_match callback. */
            $replaceCallback = function($matches) use ($variables) {
                foreach ($matches as $index => $variable) {
                    $matches[$index] = $variables[$variable] ?? $variable;
                }
                return implode('', $matches);
            };
            $value = is_string($value) ? preg_replace_callback('/__\w+__/', $replaceCallback, $value) : $value;
        };
        array_walk_recursive($content, $walkCallback);
        return $content;
    }

    /**
     * Returns a translated array from a given translation array map.
     *
     * @param array $original       The original array of words
     *                              (can be multidimensional and associative).
     * @param array $translations   The translations map. Must be associative
     *                              linking $original's value to $translation's
     *                              index.
     *
     * @return array  Returns the manufactured product.
     */
    public static function translateWords(array $original, array $translation) : array {
        /** @var callable $translate The translation callback. */
        $translate = function(&$value) use ($translation) {
            $value = $translation[$value] ?? $value;
        };
        array_walk_recursive($original, $translate);
        return $original;
    }

    /**
     * Translates the content and substitutes the variables to it's expected values.
     *
     * @param array     $content    The content to be translated and decoded.
     * @param string    $dirPath    The path where the translation can be found.
     * @param string    $language   The language to translate to.
     * @param array     $variables  The variables to replace.
     *
     * @return array    The translated and decoded content.
     */
    public static function translateAndDecode(array $content, string $dirPath, string $language, array $variables) : array {
        $translations = self::getTranslations($dirPath, $language);
        $translated = self::translateWords($content, $translations);
        return self::replaceVariables($translated, $variables);
    }
}
